/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: asgraph.cc                                   */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#include "asgraph.h"

// *************************************
// ************** AS graph *************
// *************************************

/* Warning: using this code assumes that your compiler supports partial
   expression evaluation (e.g., in (A || B), if A is true, B is not
   evaluated at all) */

/* The following are some basic tools for use with the edge weight
   type */

// Default constructor
ASEdgeWeight::ASEdgeWeight() {
   oriented = diffOriented[0] = diffOriented[1] = false;
   pathCovering[0] = pathCovering[1] = 0;
   IPCovering[0] = IPCovering[1] = 0;
   incoming = outgoing = none = undir = 0;
   graphs = snapshots = 1;
   changes = 0;
   stability = 0;
}

// Copy constructor
ASEdgeWeight::ASEdgeWeight (const ASEdgeWeight &w) {
   oriented = w.oriented;
   lastPrefix = w.lastPrefix;
   graphs = w.graphs;
   incoming = w.incoming;
   outgoing = w.outgoing;
   none = w.none;
   undir = w.undir;
   snapshots = w.snapshots;
   stability = w.stability;
   changes = w.changes;
   for (char i = 0; i <= 1; i++) {
      pathCovering[i] = w.pathCovering[i];
      IPCovering[i] = w.IPCovering[i];
      diffOriented[i] = w.diffOriented[i];
      orientation[i] = w.orientation[i];
   }
}

// Assignment operator
ASEdgeWeight& ASEdgeWeight::operator=(const ASEdgeWeight &w) {
   oriented = w.oriented;
   lastPrefix = w.lastPrefix;
   graphs = w.graphs;
   incoming = w.incoming;
   outgoing = w.outgoing;
   none = w.none;
   undir = w.undir;
   snapshots = w.snapshots;
   stability = w.stability;
   changes = w.changes;
   for (char i = 0; i <= 1; i++) {
      pathCovering[i] = w.pathCovering[i];
      IPCovering[i] = w.IPCovering[i];
      diffOriented[i] = w.diffOriented[i];
      orientation[i] = w.orientation[i];
   }
   return (*this);
}

// And here is the real code for the ASGraph class

/* The following two constants store character sequences that should
   identify the type of the graph when written to file */
const char ASGraph::graphMarker[4] = "TGR";
const char ASGraph::diffGraphMarker[4] = "TDG";

// Default constructor
ASGraph::ASGraph() {
   strcpy (error, "");
   strcpy (description, "");
   // By default, assume graph is not a differential graph
   differential = false;
}

// Destructor
ASGraph::~ASGraph() {
   while (!sourceFile.isEmpty()) {
      delete [] sourceFile.pop();
   }
}

/* This funcion converts an edge direction (IN, OUT, NONE)
   into a printable integer value (-1, 1, 0) */
int ASGraph::dirToVal(edge_dir dir) {
   switch (dir) {
      case DIR_IN:   return -1;
                     break;
      case DIR_OUT:  return 1;
                     break;
      case DIR_NONE: return 0;
                     break;
      default:       return 0;
                     break;
   }
}

/* This function converts an integer value (-1, 1, 0)
   into an edge direction (IN, OUT, NONE) */
edge_dir ASGraph::valToDir(int val) {
   switch(val) {
      case -1: return DIR_IN;
               break;
      case 1:  return DIR_OUT;
               break;
      case 0:  return DIR_NONE;
               break;
      default: return DIR_NONE;
               break;
   }
}

/* Dumps all the AS graph contents to a binary file;
   use `readFromFile' to get them back; the only
   parameter is a pointer to an open file strem;
   this method returns true if operation is successful;
   false otherwise */
bool ASGraph::writeToFile(FILE *file) {
   bool goOn, stop;
   List<graph_node*> nodeList;
   List<graph_edge*> edgeList;
   graph_node *curNode;
   graph_edge *curEdge;
   strcpy (error, "");

   // Write a string that identifies the graph
   if (!fprintf (file, "%s\n",
                 (differential ? diffGraphMarker : graphMarker))) {
      strcat (error, "Could not write file type string");
      return false;
   }

   // Write number of source files
   if (!fprintf (file, "%llu\n", sourceFile.len())) {
      strcat (error, "Could not write number of source files");
      return false;
   }

   // Write source file names
   goOn = true;               /* List of file names cannot be empty (at least
                                 one file is needed to make a graph) */
   for (sourceFile.goToFirst(); goOn; goOn = sourceFile.next()) {
      if (!fprintf (file, "%s\n", sourceFile.current())) {
         strcat (error, "Unable to write source file names");
         return false;
      }
   }

   // Write comment
   if (!fprintf (file, "%s\n", description)) {
      strcat (error, "Could not write comment");
      return false;
   }

   // Write number of nodes
   if (!fprintf (file, "%u\n", nodes())) {
      strcat (error, "Could not write number of nodes");
      return false;
   }

   // Write number of edges
   if (!fprintf (file, "%u\n", edges())) {
      strcat (error, "Could not write number of edges");
      return false;
   }

   // Write list of nodes
   nodeList = listNodes();
   stop = false;
   goOn = !nodeList.isEmpty();
   for (nodeList.goToFirst(); goOn; goOn = nodeList.next()) {
      curNode = nodeList.current();
      if (!stop && !fprintf (file, "%i\n", *(curNode->contents))) {
         strcat (error, "Could not write node");
         sprintf(error, "%s %i", error, *(curNode->contents));
         /* There is no immediate return, because that would not
            allow memory deallocation */
         stop = true;
      }
      delete curNode;
   }
   
   if (stop) {
      return false;
   }

   // Write list of edges
   edgeList = listEdges();
   stop = false;
   goOn = !edgeList.isEmpty();
   for (edgeList.goToFirst(); goOn; goOn = edgeList.next()) {
      curEdge = edgeList.current();
      if (!stop && !fprintf (file, "%i %i %i %i %lu %llu %lu %llu",
            *(curEdge->from->contents),
            *(curEdge->to->contents),
            (curEdge->weight_from->oriented ? 1 : 0),
            dirToVal(*(curEdge->orientation)),
            curEdge->weight_from->pathCovering[CMPGRAPH_FIRST],
            curEdge->weight_from->IPCovering[CMPGRAPH_FIRST],
            curEdge->weight_to->pathCovering[CMPGRAPH_FIRST],
            curEdge->weight_to->IPCovering[CMPGRAPH_FIRST])) {
         strcat (error, "Could not write edge");
         sprintf(error, "%s %i %i", error, *(curEdge->from->contents),
                                          *(curEdge->to->contents));
         /* There is no immediate return, because that would not
            allow memory deallocation */
         stop = true;
      }

      // Write edge's differential attributes
      if (differential) {
         if (!stop && !fprintf (file, " %i %i %i %i %lu %llu %lu %llu %hu %hu %hu %hu %hu %hu %f",
                 (curEdge->weight_from->diffOriented[CMPGRAPH_FIRST] ? 1 : 0),
                 (curEdge->weight_from->diffOriented[CMPGRAPH_SECOND] ? 1 : 0),
                  dirToVal(curEdge->weight_from->orientation[CMPGRAPH_FIRST]),
                  dirToVal(curEdge->weight_from->orientation[CMPGRAPH_SECOND]),
                  curEdge->weight_from->pathCovering[CMPGRAPH_SECOND],
                  curEdge->weight_from->IPCovering[CMPGRAPH_SECOND],
                  curEdge->weight_to->pathCovering[CMPGRAPH_SECOND],
                  curEdge->weight_to->IPCovering[CMPGRAPH_SECOND],
                  curEdge->weight_from->graphs,
                  curEdge->weight_from->incoming,
                  curEdge->weight_from->outgoing,
                  curEdge->weight_from->none,
                  curEdge->weight_from->undir,
                  curEdge->weight_from->changes,
                  curEdge->weight_from->stability)) {
            strcat (error, "Could not write edge");
            sprintf(error, "%s %i %i", error, *(curEdge->from->contents),
                                             *(curEdge->to->contents));
            /* There is no immediate return, because that would not
               allow memory deallocation */
            stop = true;
         }
      }
      if (!stop && !fprintf (file, "\n")) {
         strcat (error, "Could not write edge");
         sprintf(error, "%s %i %i", error, *(curEdge->from->contents),
                                          *(curEdge->to->contents));
         /* There is no immediate return, because that would not
            allow memory deallocation */
         stop = true;
      }
      delete curEdge->from;
      delete curEdge->to;
      delete curEdge;
   }

   return true;
}

/* Builds a graph from the data that have previously
   been generated by `writeToFile'; the only parameter
   is a pointer to an open file stream; returns true
   if operation is successful; false otherwise */
bool ASGraph::readFromFile(FILE *file) {
   char string[513], oldString[513];             // One more byte for '\0'
   char *err, *newFileName, *curToken;
   int intValue, sourceFiles;
   unsigned int uValue, nodeCnt, edgeCnt;
   unsigned long int ulValue;
   unsigned long long int ullValue;
   double dValue;
   ASnode from, to;
   ASEdgeWeight weight1, weight2;
   edge_dir direction;

   strcpy (error, "");

   // Get graph type
   if (!fgets(string, 513, file)) {
      strcat (error, "Could not retrieve file type string");
      return false;
   }
   // Remove CR/LF at end of line
   if (strpbrk(string, "\n\r")) {
      *(strpbrk(string, "\n\r")) = '\0';
   }
   if (strcmp(string, diffGraphMarker) == 0) {
      differential = true;
   }
   else {
      if (strcmp(string, graphMarker) == 0) {
         differential = false;
      }
      else {
         strcat (error, "Invalid file type string: `");
         strcat (error, string);
         strcat (error, "'");
         return false;
      }
   }

   // Get number of source files
   if (!fgets(string, 513, file)) {
      strcat (error, "Could not retrieve number of source files");
      return false;
   }
   // Remove CR/LF at end of line
   if (strpbrk(string, "\n\r")) {
      *(strpbrk(string, "\n\r")) = '\0';
   }
   intValue = strtol(string, &err, 10);
   if (intValue < 1 || intValue > 65536 || (strlen(err) > 0)) {
      strcat (error, "Invalid number of source files: `");
      strcat (error, string);
      strcat (error, "'");
      return false;
   }
   sourceFiles = intValue;

   // Get source files names
   for (int i = 0; i <= intValue - 1; i++) {
      if (!fgets(string, 512, file)) {
         strcat (error, "Could not retrieve all the source file names");
         return false;
      }
      // Remove CR/LF at end of line
      if (strpbrk(string, "\n\r")) {
         *(strpbrk(string, "\n\r")) = '\0';
      }
      newFileName = new char[strlen(string) + 1];
      strcpy(newFileName, string);
      sourceFile.enqueue(newFileName);
   }

   // Get comment
   if (!fgets(string, 512, file)) {
      strcat (error, "Unable to retrieve graph description string");
      return false;
   }
   // Remove CR/LF at end of line
   if (strpbrk(string, "\n\r")) {
      *(strpbrk(string, "\n\r")) = '\0';
   }
   strncpy(description, string, MAX_COMMENT_CHARS);
   // Terminate string
   description[MAX_COMMENT_CHARS] = '\0';

   // Get number of nodes
   if (!fgets(string, 512, file)) {
      strcat (error, "Could not retrieve number of nodes");
      return false;
   }
   // Remove CR/LF at end of line
   if (strpbrk(string, "\n\r")) {
      *(strpbrk(string, "\n\r")) = '\0';
   }
   uValue = strtoul(string, &err, 10);
   if (strlen(err) > 0) {
      strcat (error, "Invalid number of nodes: `");
      strcat (error, string);
      strcat (error, "'");
      return false;
   }
   nodeCnt = uValue;

   // Get number of edges
   if (!fgets(string, 512, file)) {
      strcat (error, "Could not retrieve number of edges");
      return false;
   }
   // Remove CR/LF at end of line
   if (strpbrk(string, "\n\r")) {
      *(strpbrk(string, "\n\r")) = '\0';
   }
   uValue = strtoul(string, &err, 10);
   if (strlen(err) > 0) {
      strcat (error, "Invalid number of edges: `");
      strcat (error, string);
      strcat (error, "'");
      return false;
   }
   edgeCnt = uValue;

   // Get nodes
   for (unsigned int i = 0; nodeCnt > 0 && i <= nodeCnt - 1; i++) {
      if (!fgets(string, 512, file)) {
         strcat (error, "Unable to retrieve all the nodes");
         return false;
      }
      // Remove CR/LF at end of line
      if (strpbrk(string, "\n\r")) {
         *(strpbrk(string, "\n\r")) = '\0';
      }
      intValue = strtol(string, &err, 10);
      if (intValue < 1 || intValue > 65535 || (strlen(err) > 0)) {
         strcat (error, "Invalid AS number: `");
         strcat (error, string);
         strcat (error, "'");
         return false;
      }
      addNode((ASnode) intValue, (ASnode) intValue);
   }

   // Get edges
   for (unsigned int i = 0; edgeCnt > 0 && i <= edgeCnt - 1; i++) {
      if (!fgets(string, 512, file)) {
         strcat (error, "Unable to retrieve all the edges");
         return false;
      }
      // Remove CR/LF at end of line
      if (strpbrk(string, "\n\r")) {
         *(strpbrk(string, "\n\r")) = '\0';
      }
      /* String will be tokenized, so a copy
         of it should be made */
      strcpy(oldString, string);

      // First node
      curToken = strtok(string, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      intValue = strtol(curToken, &err, 10);
      if (intValue < 1 || intValue > 65535 || (strlen(err) > 0)) {
         strcat (error, "Invalid AS number: `");
         strcat (error, curToken);
         strcat (error, "'");
         return false;
      }
      from = intValue;

      // Second node
      curToken = strtok(NULL, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      intValue = strtol(curToken, &err, 10);
      if (intValue < 1 || intValue > 65535 || (strlen(err) > 0)) {
         strcat (error, "Invalid AS number: `");
         strcat (error, curToken);
         strcat (error, "'");
         return false;
      }
      to = intValue;

      // Edge direction flag
      curToken = strtok(NULL, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      if (strcmp(curToken, "0") != 0 && strcmp(curToken, "1") != 0) {
         strcat (error, "Invalid edge direction flag: `");
         strcat (error, curToken);
         strcat (error, "'");
         return false;
      }
      weight1.oriented = weight2.oriented =
         (strcmp(curToken, "1") == 0);

      // Edge direction
      curToken = strtok(NULL, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      intValue = strtol(curToken, &err, 10);
      if (intValue < -1 || intValue > 1 || (strlen(err) > 0)) {
         strcat (error, "Invalid edge direction: `");
         strcat (error, curToken);
         strcat (error, "'");
         return false;
      }
      direction = valToDir(intValue);

      // Path covering from this side of the edge
      curToken = strtok(NULL, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      ulValue = strtoul(curToken, &err, 10);
      if (strlen(err) > 0) {
         strcat (error, "Invalid path covering value: `");
         strcat (error, curToken);
         strcat (error, "'");
         return false;
      }
      weight1.pathCovering[CMPGRAPH_FIRST] = ulValue;

      // IP address covering from this side of the edge
      curToken = strtok(NULL, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      ullValue = strtoull(curToken, &err, 10);
      if (strlen(err) > 0) {
         strcat (error, "Invalid IP address covering value: `");
         strcat (error, curToken);
         strcat (error, "'");
         return false;
      }
      weight1.IPCovering[CMPGRAPH_FIRST] = ullValue;

      // Path covering from the other side of the edge
      curToken = strtok(NULL, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      ulValue = strtoul(curToken, &err, 10);
      if (strlen(err) > 0) {
         strcat (error, "Invalid path covering value: `");
         strcat (error, curToken);
         strcat (error, "'");
         return false;
      }
      weight2.pathCovering[CMPGRAPH_FIRST] = ulValue;

      // IP address covering from the other side of the edge
      curToken = strtok(NULL, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      ullValue = strtoull(curToken, &err, 10);
      if (strlen(err) > 0) {
         strcat (error, "Invalid IP address covering value: `");
         strcat (error, curToken);
         strcat (error, "'");
         return false;
      }
      weight2.IPCovering[CMPGRAPH_FIRST] = ullValue;

      if (differential) {
         // Building back a differential graph...

         // Edge direction flag (for first compared graph)
         curToken = strtok(NULL, " ");
         if (curToken == NULL) {
            strcat (error, "Invalid edge line format: `");
            strcat (error, oldString);
            strcat (error, "'");
            return false;
         }
         if (strcmp(curToken, "0") != 0 && strcmp(curToken, "1") != 0) {
            strcat (error, "Invalid edge direction flag: `");
            strcat (error, curToken);
            strcat (error, "'");
            return false;
         }
         weight1.diffOriented[CMPGRAPH_FIRST] =
         weight2.diffOriented[CMPGRAPH_FIRST] =
            (strcmp(curToken, "1") == 0);

         // Edge direction flag (for second compared graph)
         curToken = strtok(NULL, " ");
         if (curToken == NULL) {
            strcat (error, "Invalid edge line format: `");
            strcat (error, oldString);
            strcat (error, "'");
            return false;
         }
         if (strcmp(curToken, "0") != 0 && strcmp(curToken, "1") != 0) {
            strcat (error, "Invalid edge direction flag: `");
            strcat (error, curToken);
            strcat (error, "'");
            return false;
         }
         weight1.diffOriented[CMPGRAPH_SECOND] =
         weight2.diffOriented[CMPGRAPH_SECOND] =
            (strcmp(curToken, "1") == 0);


         // Edge direction (in the first of the two compared graphs)
         curToken = strtok(NULL, " ");
         if (curToken == NULL) {
            strcat (error, "Invalid edge line format: `");
            strcat (error, oldString);
            strcat (error, "'");
            return false;
         }
         intValue = strtol(curToken, &err, 10);
         if (intValue < -1 || intValue > 1 || (strlen(err) > 0)) {
            strcat (error, "Invalid edge direction: `");
            strcat (error, curToken);
            strcat (error, "'");
            return false;
         }
         weight1.orientation[CMPGRAPH_FIRST] = valToDir(intValue);
         weight2.orientation[CMPGRAPH_FIRST] = reverse(valToDir(intValue));

         // Edge direction (in the second of the two compared graphs)
         curToken = strtok(NULL, " ");
         if (curToken == NULL) {
            strcat (error, "Invalid edge line format: `");
            strcat (error, oldString);
            strcat (error, "'");
            return false;
         }
         intValue = strtol(curToken, &err, 10);
         if (intValue < -1 || intValue > 1 || (strlen(err) > 0)) {
            strcat (error, "Invalid edge direction: `");
            strcat (error, curToken);
            strcat (error, "'");
            return false;
         }
         weight1.orientation[CMPGRAPH_SECOND] = valToDir(intValue);
         weight2.orientation[CMPGRAPH_SECOND] = reverse(valToDir(intValue));

         // Path covering from this side of the edge (for compared graph)
         curToken = strtok(NULL, " ");
         if (curToken == NULL) {
            strcat (error, "Invalid edge line format: `");
            strcat (error, oldString);
            strcat (error, "'");
            return false;
         }
         ulValue = strtoul(curToken, &err, 10);
         if (strlen(err) > 0) {
            strcat (error, "Invalid path covering value: `");
            strcat (error, curToken);
            strcat (error, "'");
            return false;
         }
         weight1.pathCovering[CMPGRAPH_SECOND] = ulValue;

         // IP address covering from this side of the edge (for compared graph)
         curToken = strtok(NULL, " ");
         if (curToken == NULL) {
            strcat (error, "Invalid edge line format: `");
            strcat (error, oldString);
            strcat (error, "'");
            return false;
         }
         ullValue = strtoull(curToken, &err, 10);
         if (strlen(err) > 0) {
            strcat (error, "Invalid IP address covering value: `");
            strcat (error, curToken);
            strcat (error, "'");
            return false;
         }
         weight1.IPCovering[CMPGRAPH_SECOND] = ullValue;

         // Path covering from the other side of the edge (for compared graph)
         curToken = strtok(NULL, " ");
         if (curToken == NULL) {
            strcat (error, "Invalid edge line format: `");
            strcat (error, oldString);
            strcat (error, "'");
            return false;
         }
         ulValue = strtoul(curToken, &err, 10);
         if (strlen(err) > 0) {
            strcat (error, "Invalid path covering value: `");
            strcat (error, curToken);
            strcat (error, "'");
            return false;
         }
         weight2.pathCovering[CMPGRAPH_SECOND] = ulValue;

         // IP address covering from the other side of the edge (for compared graph)
         curToken = strtok(NULL, " ");
         if (curToken == NULL) {
            strcat (error, "Invalid edge line format: `");
            strcat (error, oldString);
            strcat (error, "'");
            return false;
         }
         ullValue = strtoull(curToken, &err, 10);
         if (strlen(err) > 0) {
            strcat (error, "Invalid IP address covering value: `");
            strcat (error, curToken);
            strcat (error, "'");
            return false;
         }
         weight2.IPCovering[CMPGRAPH_SECOND] = ullValue;

         // ---- Differential statistics ----
         
         // Number of compared graphs
         curToken = strtok(NULL, " ");
         if (curToken == NULL) {
            strcat (error, "Invalid edge line format: `");
            strcat (error, oldString);
            strcat (error, "'");
            return false;
         }
         uValue = strtoul(curToken, &err, 10);
         if (strlen(err) > 0) {
            strcat (error, "Invalid compared graphs count value: `");
            strcat (error, curToken);
            strcat (error, "'");
            return false;
         }
         weight1.graphs = weight2.graphs = uValue;

         // Number of times edge direction was DIR_IN
         curToken = strtok(NULL, " ");
         if (curToken == NULL) {
            strcat (error, "Invalid edge line format: `");
            strcat (error, oldString);
            strcat (error, "'");
            return false;
         }
         uValue = strtoul(curToken, &err, 10);
         if (strlen(err) > 0) {
            strcat (error, "Invalid value for number of incoming edges: `");
            strcat (error, curToken);
            strcat (error, "'");
            return false;
         }
         weight1.incoming = weight2.outgoing = uValue;

         // Number of times edge direction was DIR_OUT
         curToken = strtok(NULL, " ");
         if (curToken == NULL) {
            strcat (error, "Invalid edge line format: `");
            strcat (error, oldString);
            strcat (error, "'");
            return false;
         }
         uValue = strtoul(curToken, &err, 10);
         if (strlen(err) > 0) {
            strcat (error, "Invalid value for number of outgoing edges: `");
            strcat (error, curToken);
            strcat (error, "'");
            return false;
         }
         weight1.outgoing = weight2.incoming = uValue;

         // Number of times edge direction was DIR_NONE
         curToken = strtok(NULL, " ");
         if (curToken == NULL) {
            strcat (error, "Invalid edge line format: `");
            strcat (error, oldString);
            strcat (error, "'");
            return false;
         }
         uValue = strtoul(curToken, &err, 10);
         if (strlen(err) > 0) {
            strcat (error, "Invalid value for number of undirected edges: `");
            strcat (error, curToken);
            strcat (error, "'");
            return false;
         }
         weight1.none = weight2.none = uValue;

         // Number of times edge could not be oriented
         curToken = strtok(NULL, " ");
         if (curToken == NULL) {
            strcat (error, "Invalid edge line format: `");
            strcat (error, oldString);
            strcat (error, "'");
            return false;
         }
         uValue = strtoul(curToken, &err, 10);
         if (strlen(err) > 0) {
            strcat (error, "Invalid value for number of unoriented edges: `");
            strcat (error, curToken);
            strcat (error, "'");
            return false;
         }
         weight1.undir = weight2.undir = uValue;

         // Number of times current edge's orientation has changed
         curToken = strtok(NULL, " ");
         if (curToken == NULL) {
            strcat (error, "Invalid edge line format: `");
            strcat (error, oldString);
            strcat (error, "'");
            return false;
         }
         uValue = strtoul(curToken, &err, 10);
         if (strlen(err) > 0) {
            strcat (error, "Invalid value for number of orientation changes: `");
            strcat (error, curToken);
            strcat (error, "'");
            return false;
         }
         weight1.changes = weight2.changes = uValue;
      
         // Mean duration of current edge's orientation
         curToken = strtok(NULL, " ");
         if (curToken == NULL) {
            strcat (error, "Invalid edge line format: `");
            strcat (error, oldString);
            strcat (error, "'");
            return false;
         }
         dValue = strtod(curToken, &err);
         if (strlen(err) > 0) {
            strcat (error, "Invalid value for orientation stability: `");
            strcat (error, curToken);
            strcat (error, "'");
            return false;
         }
         weight1.stability = weight2.stability = dValue;
      }

      // Add edge
      addEdge(from, to, direction, weight1, weight2);
   }
   return true;
}

/* Returns a string that describes the reason of
   last file error; this method should be invoked
   after that either one of `writeToFile' or
   `readFromFile' returned false */
char *ASGraph::errReason() {
   return error;
}

/* Returns a list of the names of the files the graph
   was generated from */
List<char*> &ASGraph::getSourceFiles() {
   return sourceFile;
}

/* Sets the graph's source file names; if the second argument
   is true, current list of source files is also cleared */
void ASGraph::setSourceFiles(char **fileName, bool clearOldList) {
   unsigned char i;
   char *currentName;
   if (clearOldList) {
      sourceFile = List<char*>();
   }
   for (i = 0; i <= 99; i++) {
      if (fileName[i] != NULL) {
         currentName = new char[strlen(fileName[i]) + 1];
         strcpy(currentName, fileName[i]);
         sourceFile.enqueue(currentName);
      }
   }
}

/* The same as above, but input is a list (instead of an array) of
   file names */
void ASGraph::setSourceFiles(List<char*> &fileList, bool clearOldList) {
   bool goOn;
   char *currentName;
   if (clearOldList) {
      sourceFile = List<char*>();
   }
   goOn = !fileList.isEmpty();
   for (fileList.goToFirst(); goOn; goOn = fileList.next()) {
      currentName = new char[strlen(fileList.current()) + 1];
      strcpy(currentName, fileList.current());
      sourceFile.enqueue(currentName);
   }
}

/* Returns the comment associated with the graph;
   an empty string if it has no comment */
char *ASGraph::getComment() {
   return description;
}

/* Sets the graph comment to a given value; if
   argument is longer than MAX_COMMENT_CHARS
   characters, it will be truncated */
void ASGraph::setComment(char *comment) {
   strncpy(description, comment, MAX_COMMENT_CHARS);
   // Terminate string
   description[MAX_COMMENT_CHARS] = '\0';
}

/* This function allows to read and/or change the
   state of the flag that marks the graph as a
   differential one; it must be set to true whenever
   differential information should be saved to file */
bool &ASGraph::isDifferential() {
   return differential;
}

// ************** AS graph *************
