/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: graph.h                                      */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#ifndef GRAPH_H
#define GRAPH_H

#include "smartstructs.h"

// ******************************************
// *****  Weighted (un)directed graph  ******
// ******************************************

enum edge_dir {DIR_NONE, DIR_IN, DIR_OUT};         /* Edge direction:
                                                      DIR_NONE - undirected edge (works like
                                                                 a double directed edge)
                                                      DIR_IN   - incoming edge
                                                      DIR_OUT  - outgoing edge */
enum edge_chng_outcome {EDGE_FLOATS,
                        EDGE_EXISTS,
                        EDGE_UPDATED,
                        EDGE_INSERTED,
                        EDGE_NOT_EXISTS,
                        EDGE_DELETED};             /* Edge insertion/deletion outcome:
                                                      EDGE_FLOATS - either of the two terminal
                                                                    nodes of the edge does not
                                                                    exist
                                                      EDGE_EXISTS - edge already exists, and
                                                                    insertion was unsuccessful
                                                      EDGE_UPDATED - edge already exists, but
                                                                     insertion/deletion caused both its
                                                                     direction and its weights
                                                                     to be updated (see below)
                                                      EDGE_INSERTED - a new edge has been inserted
                                                      EDGE_NOT_EXISTS - unable to delete edge
                                                                        because it does not exist
                                                      EDGE_DELETED - edge has successfully been deleted */

enum cc_type {FIND_SCC, FIND_WCC};                 /* Tells the method findCC whether to look for
                                                      strongly connected components (FIND_SCC) or
                                                      weakly connected components (FIND_WCC); this
                                                      difference is senseless for undirected graphs
                                                      (i.e. graphs made up of only undirected edges),
                                                      so both settings imply a search of its
                                                      connected components; anyway, using FIND_SCC
                                                      on such a graph may result in worse performance
                                                      than when using FIND_WCC */

edge_dir reverse(edge_dir orientation);            /* A method that returns the inverse
                                                      direction of an edge */

template <class nodeType, class weightType, class nodeKeyType>
   class Graph {
   protected:
      struct edge;
      struct node {
         nodeType contents;                        // Node contents
         nodeKeyType key;                          // Node key (useful in case of node listing)
         BST<edge*, nodeKeyType> neigh;            // Set of node neighbours
         char flags;                               /* This is a bit mask used during the graph
                                                      traversal */
         node *copyNode;                           /* This is used by the copy constructor and
                                                      by other methods */
      };
      struct edge {
         weightType weight_from, weight_to;        /* Weight_from is the weight of the edge
                                                      seen from the "from" side */
         edge_dir orientation;                     // Edge orientation
         node *from, *to;                          /* Quick links to the two sides of the
                                                      edge */
      };
      struct graph_node;                           // These declarations are needed for
      struct graph_edge;                           // the following variables
      unsigned int nodeCount, edgeCount;           // Keep trace of nodes and edges amount
      BST<node*, nodeKeyType> nset;                // This is the set of nodes
      List<graph_node*> nodeRetList;               // This and the following lists are used
      List<graph_edge*> edgeRetList;               // in functions that return list of items
      List<Graph*> graphRetList;

   public:
      struct graph_node {                          // This structure is both used when
         nodeType *contents;                       // reading/writing node contents and
         nodeKeyType key;                          // when returning a list of the graph
      };                                           // nodes, in order to retrieve their keys

      struct graph_edge {                          // This structure is both used when
         weightType *weight_from, *weight_to;      // reading/writing edge properties and when
         edge_dir *orientation;                    // returning a list of the graph edges
         graph_node *from, *to;
      };
      Graph();                                     // Default constructor
      ~Graph();                                    // Destructor
      Graph (const Graph&);                        // Copy constructor
      Graph& operator=(const Graph&);              // Assignment operator overload
      bool isEmpty();                              /* Returns true if the graph is empty
                                                      (neither nodes, nor edges) */
      bool addNode(const nodeType&,
                   const nodeKeyType&);            /* Adds a new (i.e. with no edges) node
                                                      to the graph; returns true if
                                                      insertion was successful; false
                                                      otherwise (node already exists) */
      graph_node* nodeContents(const nodeKeyType&);/* Gets and/or sets node contents;
                                                      also returns node key itself */
      edge_chng_outcome addEdge(
                   const nodeKeyType& source,
                   const nodeKeyType& destination,
                   const edge_dir& orientation,
                   const weightType& weight1,
                   const weightType& weight2);     /* Adds a new edge to the graph, between
                                                      two specified nodes, with a given
                                                      orientation and weight; returned value
                                                      depends on edge insertion outcome (see
                                                      above for possible values), and can be
                                                      anything but EDGE_DELETED and
                                                      EDGE_NOT_EXISTS. If an edge already
                                                      exists, and is an undirected one (DIR_NONE),
                                                      replacement with a directed one is always
                                                      successful; if existing edge is directed,
                                                      replacement with both an oppositely directed
                                                      one and an undirected one converts it
                                                      in an undirected edge; weights are assigned
                                                      regardless of edge orientation */
      graph_edge *edgeProperties(
                        const nodeKeyType& source,
                        const nodeKeyType& dest);     /* Gets and/or sets the properties of an edge
                                                         (e.g. weights, orientation); be careful:
                                                         when updating an edge by using this method,
                                                         please perform update operation twice (on
                                                         the edge itself and on its "reversed" version) */
      edge_chng_outcome updateEdge(
                   const nodeKeyType& source,
                   const nodeKeyType& destination,
                   const edge_dir& orientation,
                   const weightType& weight1,
                   const weightType& weight2);        /* A safer method for edge updating; it always
                                                         overwrites old edge attributes; returned
                                                         values can only be EDGE_NOT_EXISTS or
                                                         EDGE_UPDATED */
      List <graph_node*>& BFtraverse(
                                 const nodeKeyType&); /* Returns the list of the visited nodes
                                                         in a breadth-first traversal starting
                                                         at a specified node */
      List <graph_node*>& DFtraverse(
                                 const nodeKeyType&); /* Returns the list of the visited nodes
                                                         in a depth-first traversal starting at
                                                         a specified node */
      bool removeEdge(const nodeKeyType& source,
                      const nodeKeyType& destination);/* Removes an edge between two nodes, regardless
                                                         of its orientation; returns true if deletion
                                                         was successful; false otherwise */
      edge_chng_outcome removeEdge(
                     const nodeKeyType& source,
                     const nodeKeyType& destination,
                     edge_dir orientation);           /* Removes an edge between two nodes;
                                                         returned value depends on deletion
                                                         outcome (see above for a list of
                                                         the possible values), and can be anything
                                                         but EDGE_INSERTED, EDGE_FLOATS
                                                         and EDGE_EXISTS. If direction matches
                                                         existing edge's, it is deleted. If
                                                         attempting to remove a directed edge
                                                         while an undirected one already exists,
                                                         edge is converted in a directed one,
                                                         with an orientation that is opposite to
                                                         the original one; any other attempt will fail */
      bool removeNode(const nodeKeyType&);            /* Removes an existing node and all the
                                                         edges that weigh onto it; returns false
                                                         if node does not exist; true otherwise */
      unsigned int nodes();                           // Returns the number of nodes in the graph
      unsigned int edges();                           // Returns the number of edges in the graph
      List <graph_node*>& listNodes();                /* Returns a list of the nodes including
                                                         their key; this method should be used when
                                                         start node for a DF/BF visit is unknown */
      List <graph_edge*>& listEdges();                /* Returns a list of the edges including
                                                         their properties */
      List <graph_edge*>& listEdges(
                                 const nodeKeyType&); /* Returns a list of the edges that depart
                                                         from a specific node (regardless of their
                                                         orientation) */
      List <Graph*>& findCC(cc_type);                 /* Returns a list of the connected components
                                                         of the graph; for a list of the type of
                                                         components you can look for, see above */
};



// Default constructor
template <class nodeType, class weightType, class nodeKeyType>
   Graph<nodeType, weightType, nodeKeyType>::Graph() {
   nodeCount = edgeCount = 0;
}

// Destructor
template <class nodeType, class weightType, class nodeKeyType>
   Graph<nodeType, weightType, nodeKeyType>::~Graph() {
   List <node*> toBeVisitedN;
   List <edge*> toBeVisitedE;
   node *curNode;
   edge *curEdge;
   toBeVisitedN = nset.listItems();
   while (!toBeVisitedN.isEmpty()) {
      curNode = toBeVisitedN.pop();
      toBeVisitedE = curNode->neigh.listItems();
      while (!toBeVisitedE.isEmpty()) {
         curEdge = toBeVisitedE.pop();
         delete curEdge;
      }
      delete curNode;
   }
}

// Copy constructor
template <class nodeType, class weightType, class nodeKeyType>
   Graph<nodeType, weightType, nodeKeyType>::Graph (const Graph& source) {
   node *newNode, *curNode, *destNode;
   edge *newEdge, *curEdge;
   BST<node*, nodeKeyType> sourceNset;
   List <node*> nodeList;
   List <edge*> edgeList;
   /* Node flags here assume the following meaning:
      00000000    No flag
      00000001    Node has already been copied */
   // Passing `source' parameter as const requires working on a copy of nset
   sourceNset = source.nset;
   nodeList = sourceNset.listItems();
   nodeList.goToFirst();
   if (!nodeList.isEmpty()) {
      do {
         curNode = nodeList.current();
         if (!(curNode->flags & 0x1)) {
            // This node has not been copied yet
            curNode->flags |= 0x1;        // Mark node because it has already been copied
            destNode = new node;
            destNode->contents = curNode->contents;
            destNode->key = curNode->key;
            destNode->flags = 0x0;
            destNode->copyNode = NULL;
            curNode->copyNode = destNode;
            nset.addItem(destNode, destNode->key);
         }
         else {
            // Node already exists in destination graph: use it
            destNode = curNode->copyNode;
         }
         edgeList = curNode->neigh.listItems();
         while (!edgeList.isEmpty()) {
            curEdge = edgeList.pop();
            newEdge = new edge;
            newEdge->weight_from = curEdge->weight_from;
            newEdge->weight_to = curEdge->weight_to;
            newEdge->orientation = curEdge->orientation;
            newEdge->from = destNode;
            destNode->neigh.addItem(newEdge, curEdge->to->key);
            if (curEdge->to->flags & 0x1) {
               // "Remote" node has already been created
               newEdge->to = curEdge->to->copyNode;
               /* Edge seen from the other side will be inserted when
                  visiting `remote' node */
            }
            else {
               // Must create "remote" node
               curEdge->to->flags |= 0x1;    // Mark node as already created
               newNode = new node;
               newEdge->to = newNode;
               newNode->contents = curEdge->to->contents;
               newNode->key = curEdge->to->key;
               newNode->flags = 0x0;
               newNode->copyNode = NULL;
               curEdge->to->copyNode = newNode;
               nset.addItem(newNode, newNode->key);
            }
         }
      } while (nodeList.next());
   }
   // Now clean flags and copyNode pointer on source graph's nodes
   while (!nodeList.isEmpty()) {
      curNode = nodeList.pop();
      curNode->flags = 0x0;
      curNode->copyNode = NULL;
   }
   nodeCount = source.nodeCount;
   edgeCount = source.edgeCount;
}

// Assignment operator overload
template <class nodeType, class weightType, class nodeKeyType>
   Graph<nodeType, weightType, nodeKeyType>&
      Graph<nodeType, weightType, nodeKeyType>::operator=(const Graph& source) {
   List <node*> toBeVisitedN, nodeList, destNodeList;
   List <edge*> toBeVisitedE, edgeList;
   BST<node*, nodeKeyType> sourceNset;
   node *curNode;
   edge *curEdge;

   // Delete current object
   toBeVisitedN = nset.listItems();
   while (!toBeVisitedN.isEmpty()) {
      curNode = toBeVisitedN.pop();
      toBeVisitedE = curNode->neigh.listItems();
      while (!toBeVisitedE.isEmpty()) {
         curEdge = toBeVisitedE.pop();
         delete curEdge;
      }
      /* Deallocating memory for a node also implies releasing
         memory for the BST structure holding its edges */
      delete curNode;
   }

   // Empty existing nodes set without deleting it
   nset = BST<node*, nodeKeyType>();

   /* Using `const' modifier implies that no change can be made
      on source.nset, including invoking method listItems() */
   sourceNset = source.nset;

   // Now scan nodes and, for each one, build a copy node
   nodeList = sourceNset.listItems();
   nodeList.goToFirst();
   if (!nodeList.isEmpty()) {
      do {
         curNode = nodeList.current();
         addNode(curNode->contents, curNode->key);
      } while (nodeList.next());
   }

   // Copy edges
   while (!nodeList.isEmpty()) {
      curNode = nodeList.pop();
      edgeList = curNode->neigh.listItems();
      while (!edgeList.isEmpty()) {
         curEdge = edgeList.pop();
         /* Some of these insertions must fail, because
            using method addEdge already makes edge visible
            both from current node's side and from remote node's */
         addEdge(curEdge->from->key, curEdge->to->key,
                 curEdge->orientation, curEdge->weight_from, curEdge->weight_to);
      }
   }
   nodeCount = source.nodeCount;
   edgeCount = source.edgeCount;
   return (*this);
}

// Returns true if the graph is empty (neither nodes, nor edges)
template <class nodeType, class weightType, class nodeKeyType>
   bool Graph<nodeType, weightType, nodeKeyType>::isEmpty() {
   // Graph cannot have edges if it has no nodes
   return (nodeCount == 0);
}

/* Adds a new (i.e. with no edges) node to the graph; returns true if
   insertion was successful; false otherwise (node already exists) */
template <class nodeType, class weightType, class nodeKeyType>
   bool Graph<nodeType, weightType, nodeKeyType>::addNode(
      const nodeType& nodeContents, const nodeKeyType& nodeKey) {
   node *newNode;
   newNode = new node;
   newNode->contents = nodeContents;
   newNode->key = nodeKey;
   newNode->flags = 0x0;
   newNode->copyNode = NULL;
   if (nset.addItem(newNode, nodeKey)) {
      // This is a really new node
      nodeCount++;
      return true;
   }
   else {
      // Node already exists
      delete newNode;      // Temporary new node must be deleted
      return false;
   }
}

// Gets and/or sets node contents; also returns node key itself
template <class nodeType, class weightType, class nodeKeyType>
   Graph<nodeType, weightType, nodeKeyType>::graph_node*
      Graph<nodeType, weightType, nodeKeyType>::nodeContents(
      const nodeKeyType& nodeKey) {
   node **curNode;
   graph_node *retNode;
   curNode = nset.findItem(nodeKey);
   if (curNode!=NULL) {
      retNode = new graph_node;
      retNode->contents = &((*curNode)->contents);
      retNode->key = nodeKey;
      return retNode;
   }
   // Node could not be found
   return (graph_node*) (NULL);
}

/* Adds a new edge to the graph, between two specified nodes, with a given
   orientation and weight; returns true if insertion was successful;
   false otherwise (edge already exists). If an edge already exists, and
   is an undirected one (DIR_NONE), replacement with a directed one is
   always successful; if existing edge is directed, replacement with both
   an oppositely directed one and an undirected one converts it in an
   undirected edge; weights are assigned regardless of edge orientation */
template <class nodeType, class weightType, class nodeKeyType>
   edge_chng_outcome Graph<nodeType, weightType, nodeKeyType>::addEdge(
      const nodeKeyType& source, const nodeKeyType& destination,
      const edge_dir& orientation, const weightType& weight1,
      const weightType& weight2) {
   edge *newEdge, **curEdge, **curEdgeOtherSide;
   node **from, **to;
   from = nset.findItem(source);
   to = nset.findItem(destination);
   if (from!=NULL && to!=NULL) {
      curEdge = (*from)->neigh.findItem(destination);
      if (curEdge==NULL) {
         // Insert a new edge
         edgeCount++;
         newEdge = new edge;
         newEdge->weight_from = weight1;
         newEdge->weight_to = weight2;
         newEdge->orientation = orientation;
         newEdge->from = *from;
         newEdge->to = *to;
         (*from)->neigh.addItem(newEdge, destination);
         if (source != destination) {
            /* Self-edges don't need a copy of the
               edge representation on `remote' node */
            newEdge = new edge;
            newEdge->weight_from = weight2;
            newEdge->weight_to = weight1;
            newEdge->orientation = reverse(orientation);
            newEdge->from = *to;
            newEdge->to = *from;
            (*to)->neigh.addItem(newEdge, source);
         }
         return EDGE_INSERTED;
      }
      else {
         // Edge already exists: update it
         if (source != destination) {
            // `Conventional' edge
            curEdgeOtherSide = (*to)->neigh.findItem(source);
         }
         else {
            /* Self-edges don't want their other side updated, because
               a) they don't have one
               b) that would reverse their weight and orientation */
            curEdgeOtherSide = curEdge;
         }
         if (orientation == DIR_NONE) {
            if ((*curEdge)->orientation == DIR_NONE) {
               /* Existing edge is exactly the same as user provided:
                  abort operation */
               return EDGE_EXISTS;
            }
            else {
               /* Existing edge is directed: convert it into an
                  undirected one and update its weights */
               (*curEdgeOtherSide)->weight_from = weight2;
               (*curEdgeOtherSide)->weight_to = weight1;
               (*curEdgeOtherSide)->orientation = reverse(orientation);
               (*curEdge)->weight_from = weight1;
               (*curEdge)->weight_to = weight2;
               (*curEdge)->orientation = orientation;
               return EDGE_UPDATED;
            }
         }
         else {
            if ((*curEdge)->orientation == DIR_NONE) {
               /* Existing edge is undirected: convert it into a
                  directed one and update its weights */
               (*curEdgeOtherSide)->weight_from = weight2;
               (*curEdgeOtherSide)->weight_to = weight1;
               (*curEdgeOtherSide)->orientation = reverse(orientation);
               (*curEdge)->weight_from = weight1;
               (*curEdge)->weight_to = weight2;
               (*curEdge)->orientation = orientation;
               return EDGE_UPDATED;
            }
            else {
               // Existing edge is directed
               if ((*curEdge)->orientation == reverse(orientation)) {
                  /* Attempting to insert an edge whose orientation is opposite
                     to that of the existing one: convert the latter in an
                     undirected edge and update weights*/
                  (*curEdgeOtherSide)->weight_from = weight2;
                  (*curEdgeOtherSide)->weight_to = weight1;
                  (*curEdgeOtherSide)->orientation = DIR_NONE;
                  (*curEdge)->weight_from = weight1;
                  (*curEdge)->weight_to = weight2;
                  (*curEdge)->orientation = DIR_NONE;
                  return EDGE_UPDATED;
               }
               else {
                  /* Existing edge is exactly the same as user provided:
                     abort operation */
                  return EDGE_EXISTS;
               }
            }
         }
      }
   }
   // One of the terminal nodes of the edge does not exist: abort operation
   return EDGE_FLOATS;
}

/* Gets and/or sets the properties of an edge (e.g. weights, orientation);
   be careful: when updating an edge by using this method, please perform
   update operation twice (on the edge itself and on its "reversed"
   version) */
template <class nodeType, class weightType, class nodeKeyType>
   Graph<nodeType, weightType, nodeKeyType>::graph_edge*
      Graph<nodeType, weightType, nodeKeyType>::edgeProperties(
         const nodeKeyType& source, const nodeKeyType& destination) {
   node **sourceNode;
   graph_edge *retEdge;
   edge **curEdge;
   sourceNode = nset.findItem(source);
   if (sourceNode!=NULL) {
      curEdge = (*sourceNode)->neigh.findItem(destination);
      if (curEdge!=NULL) {
         retEdge = new graph_edge;
         retEdge->weight_from = &((*curEdge)->weight_from);
         retEdge->weight_to = &((*curEdge)->weight_to);
         retEdge->orientation = &((*curEdge)->orientation);
         retEdge->from = new graph_node;
         retEdge->from->contents = &((*sourceNode)->contents);
         retEdge->from->key = source;
         retEdge->to = new graph_node;
         retEdge->to->contents = &((*curEdge)->to->contents);
         retEdge->to->key = destination;
         return retEdge;
      }
      else {
         // Edge does not exist
         return (graph_edge*) NULL;
      }
   }
   // Soruce node does not exist
   return (graph_edge*) NULL;
}

/* A safer method for edge updating; it always overwrites old
   edge attributes; returned values can only be EDGE_NOT_EXISTS or
   EDGE_UPDATED */
template <class nodeType, class weightType, class nodeKeyType>
   edge_chng_outcome Graph<nodeType, weightType, nodeKeyType>::updateEdge(
      const nodeKeyType& source, const nodeKeyType& destination,
      const edge_dir& orientation, const weightType& weight1,
      const weightType& weight2) {
   edge **curEdge, **curEdgeOtherSide;
   node *curNode;
   curNode = nset.findItem(source);
   if (curNode != NULL) {
      curEdge = (*curNode)->neigh.findItem(destination);
      if (curEdge != NULL) {
         // If edge exists, its "remote vision" must exist too
         curEdgeOtherSide = (*curEdge)->to->neigh.findItem(source);
         (*curEdge)->orientation = orientation;
         (*curEdge)->weight_from = weight1;
         (*curEdge)->weight_to = weight2;
         (*curEdgeOtherSide)->orientation = reverse(orientation);
         (*curEdgeOtherSide)->weight_from = weight2;
         (*curEdgeOtherSide)->weight_to = weight1;
         return EDGE_UPDATED;
      }
      else {
         // Edge does not exist
         return EDGE_NOT_EXISTS;
      }
   }
   else {
      // Source node does not exist
      return EDGE_NOT_EXISTS;
   }
}

/* Returns the list of the visited nodes in a breadth-first traversal starting
   at a specified node */
template <class nodeType, class weightType, class nodeKeyType>
   List <Graph<nodeType, weightType, nodeKeyType>::graph_node*>&
   Graph<nodeType, weightType, nodeKeyType>::BFtraverse(
      const nodeKeyType& start) {
   List<node*> toBeVisited;
   List<edge*> edgeList;
   graph_node *gnode;
   node **startNode, *curNode;
   edge *curEdge;
   // Initialize an internal empty list
   nodeRetList = List<graph_node*>();
   /* Node flags are used as follows:
      00000000             No mark
      00000001             Node has already been visited */
   startNode = nset.findItem(start);
   if (startNode!=NULL) {
      toBeVisited.enqueue(*startNode);
      while (!toBeVisited.isEmpty()) {
         curNode = toBeVisited.dequeue();
         if (!(curNode->flags & 0x1)) {
            // Node has not been visited yet
            curNode->flags |= 0x1;        // Mark node as visited
            gnode = new graph_node;
            gnode->contents = &(curNode->contents);
            gnode->key = curNode->key;
            nodeRetList.enqueue(gnode);
            edgeList = curNode->neigh.listItems();
            while (!edgeList.isEmpty()) {
               // The order in which edges are considered does not matter
               curEdge = edgeList.pop();
               if (curEdge->orientation == DIR_NONE || curEdge->orientation == DIR_OUT) {
                  // Only continue visit on either outgoing or undirected edges
                  if (!(curEdge->to->flags & 0x1)) {
                     // Avoid enqueueing already visited nodes
                     toBeVisited.enqueue(curEdge->to);
                  }
               }
            }
         }
      }
   }

   // Now clear node flags
   toBeVisited = nset.listItems();
   while (!toBeVisited.isEmpty()) {
      toBeVisited.pop()->flags = 0x0;
   }

   /* Either start node does not exist, or `nodeRetList' list
      has been successfully filled */
   return (nodeRetList);
}

/* Returns the list of the visited nodes in a depth-first traversal starting at
   a specified node */
template <class nodeType, class weightType, class nodeKeyType>
   List <Graph<nodeType, weightType, nodeKeyType>::graph_node*>&
   Graph<nodeType, weightType, nodeKeyType>::DFtraverse(
      const nodeKeyType& start) {
   List<node*> toBeVisited;
   List<edge*> edgeList;
   graph_node *gnode;
   node **startNode, *curNode;
   edge *curEdge;
   // Initialize an internal empty list
   nodeRetList = List<graph_node*>();
   /* Node flags are used as follows:
      00000000             No mark
      00000001             Node has already been visited */
   startNode = nset.findItem(start);
   if (startNode!=NULL) {
      toBeVisited.push(*startNode);
      while (!toBeVisited.isEmpty()) {
         curNode = toBeVisited.pop();
         if (!(curNode->flags & 0x1)) {
            // Node has not been visited yet
            curNode->flags |= 0x1;        // Mark node as visited
            gnode = new graph_node;
            gnode->contents = &(curNode->contents);
            gnode->key = curNode->key;
            nodeRetList.enqueue(gnode);
            edgeList = curNode->neigh.listItems();
            while (!edgeList.isEmpty()) {
               // The order in which edges are considered does not matter
               curEdge = edgeList.pop();
               if (curEdge->orientation == DIR_NONE || curEdge->orientation == DIR_OUT) {
                  // Only continue visit on either outgoing or undirected edges
                  if (!(curEdge->to->flags & 0x1)) {
                     // Avoid pushing already visited nodes
                     toBeVisited.push(curEdge->to);
                  }
               }
            }
         }
      }
   }

   // Now clear node flags
   toBeVisited = nset.listItems();
   while (!toBeVisited.isEmpty()) {
      toBeVisited.pop()->flags = 0x0;
   }

   /* Either start node does not exist, or `nodeRetList' list
      has been successfully filled */
   return (nodeRetList);
}

/* Removes an edge between two nodes, regardless of its orientation;
   returns true if deletion was successful; false otherwise */
template <class nodeType, class weightType, class nodeKeyType>
   bool Graph<nodeType, weightType, nodeKeyType>::removeEdge(
      const nodeKeyType& source, const nodeKeyType& destination) {
   edge **curEdge, *curEdgeOtherSide;
   node **sourceNode;
   sourceNode = nset.findItem(source);
   if (sourceNode != NULL) {
      curEdge = (*sourceNode)->neigh.findItem(destination);
      if (curEdge != NULL) {
         edgeCount--;
         curEdgeOtherSide = *((*curEdge)->to->neigh.findItem(source));
         if (source != destination) {
            /* Delete `remote' node's vision of the edge (self-edges don't
               need such a vision) */
            delete curEdgeOtherSide;
            (*curEdge)->to->neigh.deleteItem(source);
         }
         // Delete `near' vision of the edge
         delete *curEdge;
         (*sourceNode)->neigh.deleteItem(destination);
         return true;
      }
      else {
         // Edge does not exist
         return false;
      }
   }

   // One of the terminal nodes of the edge does not exist
   return false;
}

/* Removes an edge between two nodes; returns true if removal was successful,
   false otherwise. If direction matches existing edge's, it is deleted. If
   attempting to remove a directed edge while an undirected one already exists,
   edge is converted in a directed one, with an orientation that is opposite to
   the original one */
template <class nodeType, class weightType, class nodeKeyType>
   edge_chng_outcome Graph<nodeType, weightType, nodeKeyType>::removeEdge(
      const nodeKeyType& source, const nodeKeyType& destination,
      edge_dir orientation) {
   edge **curEdge, *curEdgeOtherSide;
   node **sourceNode;
   sourceNode = nset.findItem(source);
   if (sourceNode != NULL) {
      curEdge = (*sourceNode)->neigh.findItem(destination);
      if (curEdge != NULL) {
         curEdgeOtherSide = *((*curEdge)->to->neigh.findItem(source));
         if ((*curEdge)->orientation == orientation) {
            // Existing edge's orientation matches user provided: delete edge
            edgeCount--;
            if (source != destination) {
               /* Delete `remote' node's vision of the edge (self-edges don't
                  need such a vision) */
               delete curEdgeOtherSide;
               (*curEdge)->to->neigh.deleteItem(source);
            }
            // Delete `near' vision of the edge
            delete *curEdge;
            (*sourceNode)->neigh.deleteItem(destination);
            return EDGE_DELETED;
         }
         else {
            // User provided orientation is not the same as existing edge's
            if ((*curEdge)->orientation == DIR_NONE) {
               // Update edge
               if (source != destination) {
                  curEdgeOtherSide->orientation = orientation;
               }
               (*curEdge)->orientation = reverse(orientation);
               return EDGE_UPDATED;
            }
            else {
               /* Can't delete a directed edge by trying to remove
                  an either undirected edge or an oppositely directed
                  edge */
               return EDGE_NOT_EXISTS;
            }
         }
      }
      else {
         // One of the terminal nodes of the edge does not exist
         return EDGE_NOT_EXISTS;
      }
   }

   // One of the terminal nodes of the edge does not exist
   return EDGE_NOT_EXISTS;
}


/* Removes an existing node and all the edges that weigh onto it;
   returns false if node does not exist; true otherwise */
template <class nodeType, class weightType, class nodeKeyType>
   bool Graph<nodeType, weightType, nodeKeyType>::removeNode(
      const nodeKeyType& thisNode) {
   node **sourceNode;
   List<edge*> edgeList;
   edge *curEdge, *curEdgeOtherSide;
   sourceNode = nset.findItem(thisNode);
   if (sourceNode != NULL) {
      // User provided node exists: delete it
      edgeList = (*sourceNode)->neigh.listItems();
      while (!edgeList.isEmpty()) {
         curEdge = edgeList.pop();
         if (curEdge->to->key != thisNode) {
            // This is not a self-edge
            curEdgeOtherSide = *(curEdge->to->neigh.findItem(thisNode));
            delete curEdgeOtherSide;
            curEdge->to->neigh.deleteItem(thisNode);
         }
         delete curEdge;
         /* Edge pointer in node's `neigh' field will be deleted
            when releasing memory for the node itself */
      }
      delete (*sourceNode);
      nset.deleteItem(thisNode);
      nodeCount--;
      return true;
   }

   // Node does not exist
   return false;
}

// Returns the number of nodes in the graph
template <class nodeType, class weightType, class nodeKeyType>
   unsigned int Graph<nodeType, weightType, nodeKeyType>::nodes() {
   return nodeCount;
}

// Returns the number of edges in the graph
template <class nodeType, class weightType, class nodeKeyType>
   unsigned int Graph<nodeType, weightType, nodeKeyType>::edges() {
   return edgeCount;
}

// Returns a list of the nodes
template <class nodeType, class weightType, class nodeKeyType>
   List <Graph<nodeType, weightType, nodeKeyType>::graph_node*>&
   Graph<nodeType, weightType, nodeKeyType>::listNodes() {
   node *curNode;
   List <node*> nlist;
   // Initialize an internal empty list
   nodeRetList = List<graph_node*>();
   graph_node *gnode;
   /* Retrieve all nodes, with their keys (that's why the key field
      is also placed inside each node) */
   nlist = nset.listItems();
   while (!nlist.isEmpty()) {
      curNode = nlist.pop();
      gnode = new graph_node;
      gnode->contents = &(curNode->contents);
      gnode->key = curNode->key;
      nodeRetList.enqueue(gnode);
   }
   return nodeRetList;
}

// Returns a list of the edges
template <class nodeType, class weightType, class nodeKeyType>
   List <Graph<nodeType, weightType, nodeKeyType>::graph_edge*>&
   Graph<nodeType, weightType, nodeKeyType>::listEdges() {
   edge *curEdge;
   node *curNode;
   graph_edge *gedge;
   List<node*> nodeList;
   List<edge*> edgeList;
   // Initialize an internal empty list
   edgeRetList = List<graph_edge*>();
   /* Node flags are here used as follows:
      00000000          No mark
      00000001          Node has already been visited; this implies that
                        all the edges that weigh onto it must not be
                        considered any more */
   nodeList = nset.listItems();
   nodeList.goToFirst();
   if (!nodeList.isEmpty()) {
      do {
         curNode = nodeList.current();
         edgeList = curNode->neigh.listItems();
         while (!edgeList.isEmpty()) {
            curEdge = edgeList.pop();
            if (!(curEdge->to->flags & 0x1)) {
               // Remote node has not been visited yet
               gedge = new graph_edge;
               gedge->weight_from = &(curEdge->weight_from);
               gedge->weight_to = &(curEdge->weight_to);
               gedge->orientation = &(curEdge->orientation);
               gedge->from = new graph_node;
               gedge->from->contents = &(curNode->contents);
               gedge->from->key = curNode->key;
               gedge->to = new graph_node;
               gedge->to->contents = &(curEdge->to->contents);
               gedge->to->key = curEdge->to->key;
               edgeRetList.enqueue(gedge);
            }
         }
         curNode->flags |= 0x1;           /* Mark node as visited; this can
                                             only be done after considering
                                             node's edges, because self-edges
                                             won't be listed if they meet an
                                             already visited node */
      } while (nodeList.next());
   }

   // Reset node flags
   while (!nodeList.isEmpty()) {
      nodeList.pop()->flags = 0x0;
   }

   return edgeRetList;
}

/* Returns a list of the edges that depart from a specific node
   (regardless of their orientation) */
template <class nodeType, class weightType, class nodeKeyType>
   List <Graph<nodeType, weightType, nodeKeyType>::graph_edge*>&
   Graph<nodeType, weightType, nodeKeyType>::listEdges(
      const nodeKeyType& focus) {
   node **focusNode;
   edge *curEdge;
   graph_edge *gedge;
   List<edge*> edgeList;
   // Initialize an internal empty list
   edgeRetList = List<graph_edge*>();
   focusNode = nset.findItem(focus);
   if (focusNode != NULL) {
      edgeList = (*focusNode)->neigh.listItems();
      while (!edgeList.isEmpty()) {
         curEdge = edgeList.pop();
         gedge = new graph_edge;
         gedge->weight_from = &(curEdge->weight_from);
         gedge->weight_to = &(curEdge->weight_to);
         gedge->orientation = &(curEdge->orientation);
         gedge->from = new graph_node;
         gedge->from->contents = &((*focusNode)->contents);
         gedge->from->key = (*focusNode)->key;
         gedge->to = new graph_node;
         gedge->to->contents = &(curEdge->to->contents);
         gedge->to->key = curEdge->to->key;
         edgeRetList.enqueue(gedge);
      }
   }
   /* Either node could not be found, or edge list has
      been successfully filled */
   return edgeRetList;
}

// Returns a list of the Connected Components of the graph
template <class nodeType, class weightType, class nodeKeyType>
   List <Graph<nodeType, weightType, nodeKeyType>*>&
      Graph<nodeType, weightType, nodeKeyType>::findCC(cc_type mode) {
   List<node*> toBeVisited, nodeList, visited;
   List<edge*> edgeList;
   Graph *newGraph;
   node *curNode, *startNode;
   edge *curEdge;
   bool cycle, singleNode;
   
   // Initialize an internal empty list
   graphRetList = List<Graph*>();

   /* Node flags here assume the following meaning:
      00000000             No mark
      00000001 VISITED     Node has been visited (first pass visit for SCC mode)
      00000010 INSERTED    Node has already been inserted in a CC
      00000100 REVISITED   Second pass visit flag (SCC mode only)
      00001000 ALONE       Node must be itself a SCC (SCC mode only) */

   nodeList = nset.listItems();
   nodeList.goToFirst();
   if (!nodeList.isEmpty()) {
      do {
         /* No matter in which order nodes are considered: just make sure to
            consider them all (each one must be assigned to a connected component) */
         startNode = nodeList.current();
         singleNode = false;
         if (!(startNode->flags & 0x2)) {          // INSERTED
            startNode->flags |= 0x2;               // Mark node as INSERTED
            // Prepare a new {strongly/weakly} connected component
            newGraph = new Graph;
            newGraph->addNode(startNode->contents, startNode->key);
            if (startNode->flags & 0x8) {    // ALONE
               /* A previous visit did not meet any cycle while considering this
                  node: it must be the only one in a strongly connected component */
               graphRetList.enqueue(newGraph);
               singleNode = true;
            }
            else {
               // Now start a visit (order is not relevant) from the current node
               toBeVisited.enqueue(startNode);
            }
            cycle = false;
            while (!toBeVisited.isEmpty()) {
               curNode = toBeVisited.dequeue();
               // Avoid nodes marked as ALONE (they will be places in SCCs later)
               if (!(curNode->flags & 0x8)) {
                  if (curNode->flags & 0x1) {          // VISITED
                     cycle = true;
                  }
                  else {
                     visited.enqueue(curNode);
                     curNode->flags |= 0x1;              // Mark node as VISITED
                     // Scan edges departing from that node
                     edgeList = curNode->neigh.listItems();
                     while (!edgeList.isEmpty()) {
                        curEdge = edgeList.pop();
                        if (mode == FIND_WCC) {
                           /* Weakly connected components search implies that every
                              node that is reachable from a given one over a path
                              made up of however directed edges belongs to the
                              same weakly connected component */
                           if (!(curEdge->to->flags & 0x2)) {              // INSERTED
                              // Remote node does not exist in newGraph yet
                              newGraph->addNode(curEdge->to->contents, curEdge->to->key);
                              curEdge->to->flags |= 0x2;                   // Mark remote node as INSERTED
                           }
                           if (curEdge->to == curNode) {
                              /* This is a self-edge; it cannot already exist in newGraph
                                 because each node is visited only once, and self-edges
                                 only are represented on the node they are attached to */
                              newGraph->addEdge(curNode->key, curNode->key, curEdge->orientation,
                                                curEdge->weight_from, curEdge->weight_to);
                           }
                           else {
                              // This is not a self-edge
                              if (!(curEdge->to->flags & 0x1)) {                    // VISITED
                                 toBeVisited.enqueue(curEdge->to);
                                 /* Remote node has not been visited yet; thus,
                                    this edge cannot already exist in newGraph; remember that this
                                    visit is proceeding without taking care of edge directions
                                    (which means that edges forming cycles will be inserted even
                                    without the need to cycle over nodes) */
                                 newGraph->addEdge(curNode->key, curEdge->to->key, curEdge->orientation,
                                                   curEdge->weight_from, curEdge->weight_to);
                              }
                           }
                        }
                        else {
                           // Mode is FIND_SCC: edge orientation is meaningful
                           if (curEdge->orientation != DIR_IN || curEdge->to == curNode) {
                              /* Edge is either outgoing or undirected, or it
                                 is a self-edge */
                              if (!(curEdge->to->flags & 0x1) &&                    // VISITED
                                  !(curEdge->to->flags & 0x2)) {                    // INSERTED
                                 /* Edge is either outgoing or undirected and remote node
                                    has neither been visited yet nor inserted in a CC */
                                 toBeVisited.enqueue(curEdge->to);
                              }
                              else {
                                 /* Prevent already visited nodes from being
                                    scheduled for later visit attempt: this reduces
                                    a bit the length of the toBeVisited list */
                                 cycle = true;
                              }
                           }
                        }
                     }  // while edgeList
                  }
               }
            }  // while toBeVisited
            if (mode == FIND_WCC) {
               // Weakly connected components search ends here
               graphRetList.enqueue(newGraph);
            }
            else {
               // Mode is FIND_SCC
               /* No matter whether singleNode is true or false: if it is false,
                  things will proceed as usual; if it is true, this also implies
                  that cycle is still false and visited is empty, so no operation
                  is performed */
               if (!cycle) {
                  while (!visited.isEmpty()) {
                     curNode = visited.pop();
                     // Mark all visited nodes as ALONE
                     curNode->flags |= 0x8;        // ALONE
                     // Reset visit bits
                     curNode->flags &= 0xFA;       // Mask filtering out VISITED and REVISITED
                  }
               }
               else {
                  /* This SCC is made up of more than one node; now
                     perform a visit starting at the previously considered startNode,
                     but only going over reverse-directed (i.e. incoming) edges: this
                     checks which of the visited nodes can reach startNode */
                  toBeVisited.enqueue(startNode);
                  while (!toBeVisited.isEmpty()) {
                     curNode = toBeVisited.dequeue();
                     if (!(curNode->flags & 0x4)) {      // REVISITED, ALONE
                        visited.enqueue(curNode);
                        curNode->flags |= 0x4;              // Mark node as REVISITED
                        edgeList = curNode->neigh.listItems();
                        while (!edgeList.isEmpty()) {
                           curEdge = edgeList.pop();
                           if (curEdge->orientation != DIR_OUT || curEdge->to == curNode) {
                              // Edge is either incoming or undirected, or it is a self-edge
                              if (curEdge->to->flags & 0x1) {
                                 /* Remote node could be reached from startNode (its
                                    VISITED flag is set) and can also reach startNode
                                    (this has just been discovered) */
                                 if (!(curEdge->to->flags & 0x2)) {  // INSERTED
                                    newGraph->addNode(curEdge->to->contents, curEdge->to->key);
                                    curEdge->to->flags |= 0x2;       // Mark node as INSERTED
                                 }
                                 if (curEdge->to == curNode) {
                                    /* This is a self-edge; it cannot already exist in newGraph
                                       because each node is visited only once, and self-edges
                                       only are represented on the node they are attached to */
                                    newGraph->addEdge(curNode->key, curNode->key, curEdge->orientation,
                                                      curEdge->weight_from, curEdge->weight_to);
                                 }
                                 else {
                                    // This is not a self-edge
                                    /* The edge that is about to be inserted is either a new edge or
                                       it already exists (in which case no update will take place); already
                                       existent edges are encountered because also edges forming cycles
                                       must be inserted into the SCC */
                                    newGraph->addEdge(curNode->key, curEdge->to->key, curEdge->orientation,
                                                      curEdge->weight_from, curEdge->weight_to);
                                    if (!(curEdge->to->flags & 0x4)) {                       // REVISITED
                                       toBeVisited.enqueue(curEdge->to);
                                    }
                                 }
                              }
                           }
                        }  //  while edgeList
                     }
                  }  // while toBeVisited
                  while (!visited.isEmpty()) {
                     curNode = visited.pop();
                     // Reset visit bits
                     curNode->flags &= 0xFA;       // Mask filtering out VISITED and REVISITED
                  }
               }
               // Enqueue current connected component
               if (!singleNode) {
                  graphRetList.enqueue(newGraph);
               }
            }
         }
      } while (nodeList.next());
   }

   // Now reset all the node flags
   while (!nodeList.isEmpty()) {
      nodeList.pop()->flags = 0x0;
   }

   return graphRetList;
}


// *****  Weighted (un)directed graph  ******


#endif
