/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: iptools.h                                    */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#ifndef IPTOOLS_H
#define IPTOOLS_H

#include "smartstructs.h"

// **************************************
// ************ IP Netmask **************
// **************************************

class IPNetmask {
   private:
      unsigned char bytes[4];                   // Netmask bytes
      unsigned char retBytes[4];                /* This is used when returning the four
                                                   netmask bytes */
   public:
      IPNetmask();                              // Default constructor
      ~IPNetmask();                             // Destructor
      IPNetmask(const IPNetmask&);              // Copy constructor
      IPNetmask(unsigned char,
                unsigned char,
                unsigned char,
                unsigned char);                 /* Builds a netmask from its four
                                                   bytes; illegal netmasks are
                                                   replaced with 255.255.255.255 */
      IPNetmask(char);                          /* Builds a netmask from a value
                                                   that represents the number of
                                                   bits assigned to the network */
      IPNetmask& operator=(const IPNetmask&);   // Assignment operator overload
      bool operator==(const IPNetmask&);        /* Returns true if the two objects
                                                   represent the same netmask */
      bool operator!=(const IPNetmask&);        // Inequality operator
      bool operator>(const IPNetmask&);         /* Returns true if the lvalue netmask
                                                   has more `1's than the other one */
      bool operator>=(const IPNetmask&);        /* Greater than or equal to operator
                                                   overload */
      bool operator<(const IPNetmask&);         /* Returns true if the lvalue netmask
                                                   has less `1's than the other one */
      bool operator<=(const IPNetmask&);        // Less than or equal to operator overload
      void setNetmask(unsigned char,
                      unsigned char,
                      unsigned char,
                      unsigned char);           /* Changes the netmask by setting its
                                                   four bytes; illegal netmasks are
                                                   replaced with 255.255.255.255 */
      void setNetmask(char netBits);            /* Changes the netmask on the basis
                                                   of its `brief' value (i.e., the
                                                   number of bytes that define the
                                                   network) */
      unsigned char *getNetmask();              /* Returns the four bytes of the
                                                   netmask; warning: subsequent callings
                                                   to this function always return the
                                                   same address, so avoid storing it
                                                   for later use because the bytes it
                                                   points to can be altered by a later
                                                   call to this method */
      char brief();                             /* Returns the netmask in the `brief'
                                                   format (i.e. as the number of
                                                   bits that form the network) */
};

// ************ IP Netmask **************







// **************************************
// ************ IP Address **************
// **************************************

class IPAddress {
   private:
      unsigned char addrByte[4];                /* The four address bytes, the first
                                                   being the most significant:
                                                   addrByte[0].addrByte[1].addrByte[2].addrByte[3] */
      unsigned char retAddrByte[4];             /* This is used when returning the four
                                                   address bytes */
   public:
      IPAddress();                              // Default constructor
      ~IPAddress();                             // Destructor
      IPAddress(unsigned char,                  // Another constructor
                unsigned char,
                unsigned char,
                unsigned char);
      IPAddress(const IPAddress&);              // Copy constructor
      IPAddress& operator=(const IPAddress&);   // Assignment operator overload
      /* The following operators compare two addresses with respect to their
         "position" in the IP address space */
      bool operator==(const IPAddress&);        // Equality test overload
      bool operator!=(const IPAddress&);        // Inequality test overload
      bool operator>(const IPAddress&);         /* Greater than operator overload;
                                                   ordering of IP addresses is based on
                                                   the value of their bytes (the leftmost
                                                   being the most significant) */
      bool operator>=(const IPAddress&);        // Greater than or equal to operator overload
      bool operator<(const IPAddress&);         // Less than operator overload
      bool operator<=(const IPAddress&);        // Less than or equal to operator overload

      IPAddress operator+(
                  const unsigned long long int&); /* Plus sign overload; adding an integer value n
                                                   to an address results in the n-th address that
                                                   follows the given one in the IP address space*/
      IPAddress& operator+=(
                  const unsigned long long int&); /* Increment assignment operator overload; works the
                                                   same as a=a+n, where a is an IPAddress and n a long int */
      unsigned long long int operator-(
                     const IPAddress&);         /* Minus sign overload; the difference between two IP
                                                   addresses is the number of addresses in the IP space
                                                   that fall between the two given ones (including themselves);
                                                   i.e., the minimum value that this operator returns is 1
                                                   (when applying it between an address and itself);
                                                   returned result is positive, whichever the ordering of
                                                   the two addresses is */
      IPAddress operator-(
               const unsigned long long int&);  /* Minus sign overload; subtracting an integer value n
                                                   from an address results in the n-th address that
                                                   precedes the given one */
      IPAddress& operator-=(
               const unsigned long long int&);  /* Decrement assignment operator overload; works the
                                                   same as a=a-n, where a is an IPAddress and n a long int */
      char ipClass();                           /* Returns the address class:
                                                   'A' for A class (/8)
                                                   'B' for B class (/16)
                                                   'C' for C class (/24)
                                                   'D' for D class (multicast)
                                                   'E' for E class (future use)
                                                   'U' for unknown class */
      IPAddress mask(const IPNetmask&);         /* Computes the bitwise "and" of the current
                                                   address with a mask; does not alter current
                                                   address */
      unsigned char *getAddress();              /* Returns the four address bytes; warning:
                                                   subsequent callings to this function
                                                   always return the same address, so
                                                   avoid storing it for later use because
                                                   the bytes it points to can be altered
                                                   by a later call to this method */
      void setAddress(unsigned char,            // Sets the four address bytes
                      unsigned char,
                      unsigned char,
                      unsigned char);
};

// ************* IP Address **************






// ***************************************
// ************** IP Prefix **************
// ***************************************

class IPPrefix {
   private:
      IPAddress addr;                           // Prefix address
      IPNetmask mask;                           // Prefix netmask
      List<IPAddress> addrRetList;              /* This list is used when returning the list
                                                   of addresses that a prefix represents */
   public:
      IPPrefix();                               // Default constructor
      ~IPPrefix();                              // Destructor
      IPPrefix(const IPAddress&);               /* Builds a prefix using only
                                                   an IP address; netmask is
                                                   based upon address class */
      IPPrefix(const IPAddress&,
               const IPNetmask&);               /* Typical definition of a prefix:
                                                   address and netmask; a check
                                                   is performed to ensure
                                                   address matches the beginning
                                                   of the address block implied
                                                   by the netmask */
      IPPrefix& operator=(const IPPrefix&);     // Assignment operator overload
      bool operator==(const IPPrefix&);         /* Returns true if the two objects
                                                   represent the same prefix (i.e.,
                                                   they have the same address and netmask) */
      bool operator!=(const IPPrefix&);         // Inequality operator overload
      bool isMoreSpecific(const IPPrefix&);     /* Returns true only when current
                                                   object represents a prefix
                                                   that is more specific than
                                                   a given one */
      bool isLessSpecific(const IPPrefix&);     /* The same as `isMoreSpecific',
                                                   but in reverse order */

      IPAddress address();                      /* Returns the address field of the
                                                   prefix */
      IPNetmask netmask();                      // Returns the netmask of the prefix
      unsigned long long int addresses();       /* Returns the number of addresses
                                                   in the prefix */
      List<IPAddress>& listAddresses();         /* Returns a list of the addresses
                                                   that the prefix summarizes */
};

// ************** IP Prefix **************






// **************************************
// ************  IP Subspace  ***********
// **************************************

class IPSubSpace {
   private:
      struct IPInterval {
         IPAddress begin, end;
      };
      unsigned long long int addrCount;            // Traces the number of addresses in the set
      BST<IPInterval, IPAddress> addrSet;          /* Stores the whole set of addresses (i.e.
                                                      those that form the subspace) */
      List<IPAddress> addrRetList;                 // This variable and the following are used
      List<IPPrefix> prefRetList;                  // when returning a list of addresses/prefixes
                                                   // stored in the set
   public:
      IPSubSpace();                                // Default constructor; builds an empty set
      ~IPSubSpace();                               // Destructor
      bool isEmpty();                              /* Returns true if the current object represents
                                                      the empty subspace (i.e. no address) */
      IPSubSpace(const IPSubSpace&);               // Copy constructor
      IPSubSpace(const IPPrefix&);                 /* Another constructor; initializes the set
                                                      with a given prefix (address, netmask) */
      IPSubSpace(const IPAddress&);                /* Like the previous constructor, but without a
                                                      netmask */
      IPSubSpace& operator=(const IPSubSpace&);    // Assignment operator overload
      bool operator==(const IPSubSpace&);          /* Tests whether the two subspaces hold the
                                                      same addresses */
      bool operator!=(const IPSubSpace&);          // Inequality operator overload
      bool includes(const IPSubSpace&);            /* Returns true if current subspace includes
                                                      a given one (strict inclusion) */
      bool isIncluded(const IPSubSpace&);          /* Returns true if current subspace is included
                                                      in a given one (strict inclusion) */
      IPSubSpace intersect(const IPSubSpace&);     // Computes the intersection of two address sets
      IPSubSpace& operator*(const IPSubSpace&);    // Alias for `intersect'
      IPSubSpace& operator*=(const IPSubSpace&);   // *= operator overload
      IPSubSpace join(const IPSubSpace&);          // Computes the union of two address sets
      IPSubSpace& operator+(const IPSubSpace&);    // Alias for `join'
      IPSubSpace& operator+=(const IPSubSpace&);   // += operator overload
      unsigned long long int addresses();          /* Returns the number of addresses that currently
                                                      are in the set */
      List<IPAddress>& listAddresses();            // Returns a list of the addresses in the set
      List<IPPrefix>& listPrefixes();              /* Returns a list of the addresses in the set, in
                                                      the form address/netmask (IP prefix); performs
                                                      something similar to the CIDR aggregation */
};

// ************  IP Subspace  ***********


#endif
