/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: tdiffGraph.h                                 */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#ifndef TDIFFGRAPH_H
#define TDIFFGRAPH_H

#include "list.h"
#include "asgraph.h"
#include "consoletools.h"
#include <stdio.h>
#include <errno.h>

/* These are the edge properties that can be looked for when
   building an overlap graph. Each one means that an
   edge in the overlap graph will be inserted if... */
enum diffGraph_type {OVER_ORIENT1ST = 0x1,      // ...edge is oriented in the 1st graph
                     OVER_ORIENT2ND = 0x2,      // ...edge is oriented in the 2nd graph
                     OVER_OPPDIRECTED = 0x4,    // ...edge has opposite directions in the two graphs
                     OVER_DIFFDIRECTED = 0x8,   // ...edge has different directions in the two graphs
                     OVER_PEER1ST = 0x10,       // ...edge is a peering in the 1st graph
                     OVER_PEER2ND = 0x20,       // ...edge is a peering in the 2nd graph
                     OVER_NONPEER1ST = 0x40,    // ...edge is not a peering in the 1st graph
                     OVER_NONPEER2ND = 0x80,    // ...edge is not a peering in the 2nd graph
                     OVER_UNORIENT1ST = 0x100,  // ...edge is not oriented in the 1st graph
                     OVER_UNORIENT2ND = 0x200,  // ...edge is not oriented in the 2nd graph
                     OVER_WEIGHT = 0x400,       // ...any of the edge weights in the two graphs differ
                     OVER_ANY = 0x800};         // ...the edge just exists in both graphs

/* These are the edge properties that can be looked for when
   building a complementary graph; each one means that an
   edge will be inserted in the complementary graph if... */
enum compGraph_type {COMP_UNORIENTED1ST = 0x1,  // ...it exists only and is unoriented in the 1st graph
                     COMP_PEERING1ST = 0x2,     // ...it exists only and is a peering edge in the 1st graph
                     COMP_DIRECTED1ST = 0x4,    // ...it exists only and is a directed edge in the 1st graph
                     COMP_ORIENTED1ST = 0x8,    // ...it exists only and is an oriented edge in the 1st graph
                     COMP_ANY1ST = 0x10,        // ...it exists only in the 1st graph
                     COMP_UNORIENTED2ND = 0x20, // The same as above, but for the 2nd graph
                     COMP_PEERING2ND = 0x40,
                     COMP_DIRECTED2ND = 0x80,
                     COMP_ORIENTED2ND = 0x100,
                     COMP_ANY2ND = 0x200};

/* This structure contains data about numerical differences
   between two graphs */
struct diff_type {
   unsigned int nodesOnlyInThe1stGraph,      // Number of nodes that are present
                nodesOnlyInThe2ndGraph,      // in only one of the graphs or
                commonNodes,                 // in both of them.
                edgesOnlyInThe1stGraph,      // The same for the edges
                edgesOnlyInThe2ndGraph,
                commonEdges,
                peersIn1stGraph,             // Number of peerings in the 1st graph
                peersIn2ndGraph,             // Number of peerings in the 2nd graph
                directedIn1stGraph,          // Number of non-peerings in the 1st graph
                directedIn2ndGraph,          // Number of non-peerings in the 2nd graph
                unorientedIn1stGraph,        // Number of unoriented edges in the 1st graph
                unorientedIn2ndGraph,        // Number of unoriented edges in the 2nd graph
                diffDirected,                // Number of differently directed edges
                oppDirected,                 // Number of oppositely directed edges
                consDirected,                // Number of consistently directed edges
                peersOnlyInThe1stGraph,      // Number of peers among edges only in the 1st graph
                peersOnlyInThe2ndGraph,      // Number of peers among edges only in the 2nd graph
                directedOnlyInThe1stGraph,   // Number of non-peers among edges only in the 1st graph
                directedOnlyInThe2ndGraph,   // Number of non-peers among edges only in the 2nd graph
                unorientedOnlyInThe1stGraph, // Number of unoriented edges among those only in the 1st graph
                unorientedOnlyInThe2ndGraph; // Number of unoriented edges among those only in the 2nd graph
};

// This structure holds parameters got from the command line
struct cmdLineParameters {
   cmdLineParameters() {
      outFile = NULL;
      strcpy(comment, "");
   }
   ~cmdLineParameters() {
      while (!graphFileName.isEmpty()) {
         delete graphFileName.pop();
      }
      if (outFile != NULL) {
         delete outFile;
      }
   }

   List<int> mode;               /* This is a list of clauses (the list itself
                                    represents a formula in DNF) which estabilish
                                    the criteria of edge insertion when producing
                                    an overlap graph */
   List<int> cmpMode;            /* This is another list of conditions that represents
                                    the conditions that edges must obey when building
                                    a complementary graph */
   List<char*> graphFileName;    // List of user provided source file names
   char *outFile,                /* Name of the file to write statistics about
                                    differences in */
        *graphFile,              // Name of the file to write the graph in
        comment[513];            /* An optional description string for the graph
                                    file; by default, graph is always labelled with
                                    the names of the files it was generated from */
   bool listNodes,               /* When this is true, instead of only returning
                                    numerical statistics, the program outputs lists
                                    of nodes that are only in one of the graphs or
                                    in both of them */
        negate,                  /* If this is true, the condition represented by
                                    the `mode' or `cmpMode' list must be considered
                                    as negated */
        diffGraph,               /* User asked to build an overlap or
                                    complementary graph */
        multiGraph,              /* User asked to make a comparison among more
                                    than 2 graphs */
        force,                   /* When this is true, existing files are always
                                    overwritten */
        help;                    // When true, a command line help is shown
};

// This function just prints out program name
void printHeader(FILE*);

/* This function processes command line parameters and returns
   true if no error is found; false if there is any syntax error */
bool parseArguments(int argc, char **argv, cmdLineParameters &params, char *error);

// This function prints an on line help
void printHelp(FILE*, char *cmdLine);

/* This function compares two edges according to their
   "lexicographic" ordering (of their terminal nodes),
   and returns a value that is less than, equal to or
   more than 0 if, respectively, first edge is found
   to be "smaller", "equal to" or "greater" than the
   second one */
int lexOrder(ASGraph::graph_edge*, ASGraph::graph_edge*);

/* This function compares nodes in the two graphs and returns
   how many of them are present in only one or both of them; if
   asked for, it also prints a list of the nodes that satisfy
   such a property */
void nodeCmp(FILE *file, cmdLineParameters &params, ASGraph &graph1,
             ASGraph &graph2, diff_type &results);

/* This function updates mean orientation durations on
   a multi-differential graph; it should be used after
   processing the last graph */
void updateMeans(ASGraph&);

/* This function alters an existing 'base' (differential) graph by updating
   its "statistic" fields with information coming from other graphs; the
   function should be invoked for each graph that must be compared, but
   always using the same 'base' graph */
void graphCompare(ASGraph &baseGraph, ASGraph &graph);

/* This function actually inserts an edge inside an
   overlap graph; it retrieves edge properties
   seen on the two compared graphs from the first
   two arguments */
void addDiffGraphEdge(ASGraph::graph_edge *edge1, ASGraph::graph_edge *edge2,
                      ASGraph &diffGraph);

/* This function checks whether an edge conforms to
   a user provided set of conditions; it must be applied
   when building a complementary graph */
bool isCmpGraphEdge(ASGraph::graph_edge *edge, ASGraph &diffGraph,
                    List<int> &conds, bool leftGraph);

/* This function checks whether a given couple of
   edges matches user provided conditions (i.e.,
   it tests whether an edge on the overlap graph
   should actually be inserted) */
bool isDiffGraphEdge(ASGraph::graph_edge *edge1, ASGraph::graph_edge *edge2,
                     ASGraph &diffGraph, List<int> &clauses);

/* This function accomplishes two operations: it computes numerical differences
   between the edges of the two graphs and it builds a graph corresponding
   to these differences */
void makeDiffGraph(FILE *file, cmdLineParameters &params,
                   ASGraph &graph1, ASGraph &graph2, ASGraph &diffGraph,
                   diff_type &results);

#endif
