/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: tmakeGraph.h                                 */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#ifndef TMAKEGRAPH_H
#define TMAKEGRAPH_H

#include "asgraph.h"
#include "fileparsers.h"
#include "consoletools.h"
#include <string.h>
#include <stdio.h>
#include <errno.h>

// A list of accepted file types
enum file_type {FILETYPE_BGPDUMP,      // File is a BGP (`show ip bgp') dump
                FILETYPE_SAT,          // File is the output of a bgpPathSat computation
                FILETYPE_VANTAGE,      // File is relationship file provided by Agarwal et al.
                FILETYPE_PATH,         // File is a list of AS paths
                FILETYPE_UNKNOWN};     /* File is a list of AS pairs for which an orientation
                                          could not be inferred */

// A list of possible file contents
enum file_contents {FILECONTENTS_NODES = 0x1,            // Get graph nodes from this file
                    FILECONTENTS_EDGES = 0x2,            /* Get graph edges (just edges, with no
                                                            direction) from this file */
                    FILECONTENTS_EDGEDIRECTION = 0x4,    /* Get graph edges directions from this
                                                            file, and apply them only to
                                                            existing edges */
                    FILECONTENTS_PATHCOVERING = 0x8,     // Get path covering from this file
                    FILECONTENTS_PREFIXCOVERING = 0x10}; // Get IP address covering from this file

// The following structure stores command line parameters
struct cmdLineParams {
   cmdLineParams() {
      strcpy(comment, "");
      for (int i = 0; i <= 99; i++) {
         fileName[i] = NULL;
         prependAS[i] = 0;         // Dummy value
      }
      outFile = NULL;
      graphFile = NULL;
   }

   ~cmdLineParams() {
      for (int i = 0; i <= 99; i++) {
         if (fileName[i] != NULL) {
            delete fileName[i];
         }
      }
      if (outFile != NULL) {
         delete [] outFile;
      }
      if (graphFile != NULL) {
         delete [] graphFile;
      }
   }

   char *fileName[100];                   // Input file names will be stored here
   file_type fileType[100];               // File types for input files will be stored here
   int fileContents[100];                 // File contents for input files will be stored here
   ASnode prependAS[100];                 /* Number of the AS that should be prepended to
                                             every AS path in the BGP dump; 0 if no
                                             AS should be prepended */
   short int nonBGPfiles, BGPfiles;       /* Total number of non-BGP-dump-files and
                                             BGP-dump-files from command line */
   bool verbose;                          /* When verbose mode is used, any parsing error
                                             is reported */
   bool force;                            /* When this flag is set to true, files will
                                             be automatically overwritten without prompting
                                             user before */
   char comment[MAX_COMMENT_CHARS + 1];   /* This is a string that holds comments for
                                             current graph */
   char *outFile;                         /* The name of the output file (i.e., the file
                                             the graph will be written to) */
   char *graphFile;                       /* An optional name of a graph that should be
                                             refined with new data */
   bool showHelp;                         /* This is set to true when user asks for command
                                             line help */
};

/* This function swaps data about file names, contents and types
   inside a given cmdLineParams structure */
void dynaSwap(cmdLineParams &params, int index1, int index2);

/* This function parses command line parameters and writes them
   inside a structure */
bool parseArguments (int argc, char **argv, cmdLineParams &params, char *error);

// This function just prints the name of the tool
void printHeader(FILE *);

/* This function checks whether there is at least a true element in
   an array of boolean elements */
bool anyTrue (bool array[], int howManyElements);

// This function prints out a command line help
void printHelp(FILE*);

/* This function builds an internal representation of the desired graph,
   based on an array of user-provided file names (`fileName'); please
   ensure it stores file names in the order in which they must be actually
   read (i.e., the first file is the one to get nodes from, etc.; BGP
   dump files must be the last ones); returned value is 0 upon success;
   non-zero if a fail occurs */
int buildGraph(short int nonBGPfiles, short int BGPfiles, char **fileName,
               file_type *fileType, int *fileContents,
               bool showPercentage, bool verbose, ASnode *prependAS, ASGraph &graph);

#endif
